/*
 * a1-l.u.c
 * Copyright (C) 2011-2017, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>
#include <string.h>

#include <a1-copy.h>
#include <a1-inter.h>
#include <a1-types.h>
#include <e4.h>
#include <lxcall.h>
#include <lxcast.h>

#define POST_BITS \
    X1f4_LXCALL_POST_ARGS | X1f4_LXCALL_POST_DATA | X1f4_LXCALL_POST_LINE

#define I_MODE(t)			(*((X1f4_E4_C_MODE *) (t)))
#define I_USER(t)			(*((X1f4_E4_C_USER *) (t)))

#define l_USER(e, output) \
    {									      \
	X1f4_E4_C_USER *l;						      \
									      \
	l = (void *) (output);						      \
	*l = (e);							      \
    }

typedef struct pick_type {
    struct x1f4_dxcast_type dxcast;
    struct x1f4_function_type function;
    struct x1f4_linetext_type linetext;
} pick_type;

typedef struct slip_type {
    int (*call)(void *, int, void *, struct x1f4_dispatch_type **), lock, miss,
	slip, tail;
    struct a1_instant_type *instant_data;
    struct a1_shuffle_type *shuffle_data;
    struct x1f4_dxcast_type dxcast;
    struct x1f4_function_type function;
    const struct x1f4_function_type *function_data, *function_text;
    struct x1f4_linetext_type linetext, *linetext_data;
    unsigned class, flags, parts;
    void *back, *clear[3], **track;
} slip_type;

extern const void *const x1f4_a1_walk_link;

static int beta_call(void *, void *, void **);
static int ever_call(void **, void **, unsigned, unsigned, char *, int *,
		     struct a1_instant_type *, const int *, const int *,
		     unsigned, int, unsigned, int);
static int fine_call(struct a1_instant_type *, void *, void **, void **,
		     struct x1f4_linetext_type *,
		     const struct x1f4_function_type *, int, int,
		     const struct x1f4_function_type *, int);
static int part_call(void *, int, void *, struct x1f4_dispatch_type **);

static const int c_____c_m[] = {
/* *INDENT-OFF* */
    X1f4_E4_CASE,
    X1f4_E4_MODE
/* *INDENT-ON* */
};
const struct x1f4_function_type _x1f4_a1_l_u = {
/* *INDENT-OFF* */
    "ucall",
    beta_call,
    X1f4_E4_CASE,
    c_____c_m,
    2,
    X1f4_E4_KEEP_CALL | X1f4_E4_LINK_PASS | X1f4_E4_SIDE_LIST
    | X1f4_E4_WALK_LINK,
    5
/* *INDENT-ON* */
};

static int
beta_call(void *context, void *output, void **input)
{
    int status;
    struct a1_instant_type *instant_data;
    struct x1f4_dxcast_type *dxcast_data;
    void **track, **used;

    if (output == x1f4_a1_walk_link) {
	track = input[1];
    } else {
	track = input;
    }

    used = track[-1];

    instant_data = used[X1f4_E4_USED_META];

    dxcast_data = I_USER(track[0]);

    if (X1f4_E4_CALL < dxcast_data->type) {
	struct x1f4_linetext_type *linetext_data;

	linetext_data = I_USER(&dxcast_data->data);
	if (linetext_data->function.function) {
	    X1f4_E4_C_MODE lock;
	    const struct x1f4_function_type *function_text;
	    unsigned class;

	    lock = I_MODE(track[1]);

	    function_text = used[X1f4_E4_RENDITION];

	    class = function_text->count - 2 + 1;
	    if (lock < -(int) class || !(lock < (int) class)) {
		status = _x1f4_a1_stat_land
		    (instant_data->shuffle_data, lock, class);
	    } else {
		const struct x1f4_function_type *function_data;

		function_data = &instant_data->fail_node->linetext.function;

		if (function_data->flags
		    & (X1f4_E4_SIDE_LIST | X1f4_E4_SLIP_LIST)) {
		    function_data = instant_data->s.function_data;
		}

		if (function_data->count) {
		    status = fine_call
			(instant_data, output, input, track, linetext_data,
			 function_data, 0, -1, function_text, lock);
		} else {
		    status = 0;

		    if (output == x1f4_a1_walk_link) {
			input[2] = NULL;
		    }
		}
	    }
	} else {
	    status =
		_x1f4_a1_stat_zero(instant_data->shuffle_data, "ucall", 5);
	}
    } else {
	status = _x1f4_a1_stat_call(instant_data->shuffle_data, "ucall", 5);
    }

    return status;
}


static int
ever_call(void **post_data, void **line_data, unsigned parts, unsigned flags,
	  char *used, int *post_args, struct a1_instant_type *instant_data,
	  const int *line_args, const int *link_args, unsigned line, int lock,
	  unsigned link, int call)
{
    struct a1_link_type *link_data;
    struct a1_node_type *node_data;
    struct x1f4_dxcast_type *object_text;
    unsigned class, post_type;
    void **rail, *slip;

    node_data = instant_data->fail_node;

    link_data = node_data->link_data;

    line_data += 2;

    memcpy(post_data, line_data, lock * sizeof(void *));

    post_data += lock;

    class = node_data->linetext.function.count;

    object_text = instant_data->s.lead;
    object_text -= class;

    rail = (void *) (object_text + link);
    rail -= class;

    link_args += call;

    post_type = flags & X1f4_E4_POST_TYPE;

    if (call < class) {
	slip = used + (integral_q) link_data[call].datatext.link;
	if (post_type) {
	    if (link_args[link] & X1f4_E4_POST_XSET) {
		slip = *(void **) slip;
	    }
	}
    } else {
	if (post_type) {
	    if (link_args[link] & X1f4_E4_POST_XSET) {
		slip = rail[call];
	    } else {
		if (*link_args == X1f4_E4_CASE) {
		    rail[call] = object_text + call;
		    slip = rail + call;
		} else {
		    slip = &object_text[call].data;
		}
	    }
	} else {
	    if (*link_args == X1f4_E4_CASE) {
		rail[call] = object_text + call;
		slip = rail + call;
	    } else {
		slip = &object_text[call].data;
	    }
	}
    }

    *post_data++ = slip;

    memcpy(post_data, line_data + lock, (line - lock) * sizeof(void *));

    line_args += 2;

    memcpy(post_args, line_args, lock * sizeof(int));
    post_args += lock;

    *post_args++ = *link_args;

    memcpy(post_args, line_args + lock, (line - lock) * sizeof(int));
    post_args += line - lock;

    if (post_type | (parts & X1f4_E4_POST_TYPE)) {
	line_args += line + 2;

	memcpy(post_args, line_args, lock * sizeof(int));
	post_args += lock;

	if (post_type) {
	    *post_args = link_args[link];
	} else {
	    *post_args = 0;
	}

	post_args++;

	memcpy(post_args, line_args + lock, (line - lock) * sizeof(int));
    }

    return 0;
}


static int
fine_call(struct a1_instant_type *instant_data, void *output, void **input,
	  void **track, struct x1f4_linetext_type *linetext_data,
	  const struct x1f4_function_type *function_data, int head, int tail,
	  const struct x1f4_function_type *function_text, int lock)
{
    int linked, status;
    unsigned class, count, flags, grade, parts, scale = 0;
    void *beta;

    flags = function_data->flags;

    count = function_data->count;

    if (head < 0) {
	head += count;
    }
    if (tail < 0) {
	tail += count;
    }

    class = function_text->count - 2;

    grade = class + 1;

    if (lock < 0) {
	lock += grade;
    }

    parts = function_text->flags;

    if ((flags | parts) & X1f4_E4_POST_TYPE) {
	scale = 1;
    }

    linked = output == x1f4_a1_walk_link
	? sizeof(struct slip_type) : sizeof(struct pick_type);

    status = instant_data->v.link
	(instant_data->v.data, &beta,
	 linked + (sizeof(void *) + (sizeof(int) << scale)) * grade
	 + sizeof(void *));
    if (status) {
	status = _x1f4_a1_stat_link(instant_data->shuffle_data);
    } else {
	struct a1_shuffle_type *shuffle_data;
	struct x1f4_function_type *function_call;
	void **frame;

	if (output == x1f4_a1_walk_link) {
	    function_call = &((struct slip_type *) beta)->function;
	} else {
	    function_call = &((struct pick_type *) beta)->function;
	}

	function_call->count = grade;

	function_call->flags = (flags | parts) & X1f4_E4_POST_TYPE;

	frame = (void *) ((char *) beta + linked);
	frame++;

	function_call->args = (void *) (frame + grade);

	frame[-1] = track[-1];

	tail++;

	shuffle_data = instant_data->shuffle_data;

	if (output == x1f4_a1_walk_link) {
	    int type;
	    struct slip_type *call_data;
	    void **clear, *series;

	    type = linetext_data->function.type;

	    call_data = beta;

	    series = input[0];

	    if (type != X1f4_E4_CASE) {
		l_USER(&call_data->dxcast, series);

		call_data->dxcast.call = 0;
		call_data->dxcast.type = type;

		series = &call_data->dxcast.data;
	    } else {
	    }

	    ever_call
		(frame, track, parts, flags, track[-1],
		 (int *) function_call->args,
		 instant_data, function_text->args, function_data->args, class,
		 lock, count, head);

	    clear = call_data->clear;

	    clear[0] = series;
	    clear[1] = frame;

	    call_data->linetext = *linetext_data;

	    call_data->linetext.function.type = X1f4_E4_CASE;

	    status = _libx1f4i0_lxcall_post_call
		(&call_data->linetext, function_call, output, 0, clear,
		 POST_BITS, &instant_data->v,
		 &shuffle_data->p, &shuffle_data->q, &shuffle_data->x,
		 shuffle_data->y.eelookup_data);
	    if (status) {
	    } else {
		struct x1f4_dispatch_type *dispatch_data;

		dispatch_data = input[2] = clear[2];

		if (dispatch_data) {
		    call_data->back = dispatch_data->back;
		    call_data->call = dispatch_data->call;
		    call_data->lock = lock;
		    call_data->miss = count;
		    call_data->slip = head;
		    call_data->tail = tail;

		    call_data->class = class;
		    call_data->flags = flags;
		    call_data->parts = parts;
		    call_data->track = track;

		    call_data->instant_data = instant_data;
		    call_data->shuffle_data = shuffle_data;

		    call_data->function_data = function_data;
		    call_data->function_text = function_text;
		    call_data->linetext_data = linetext_data;

		    dispatch_data->back = call_data;
		    dispatch_data->call = part_call;

		    call_data->linetext.function.type = type;
		}
	    }
	} else {
	    int call, type;
	    struct pick_type *call_data;

	    type = linetext_data->function.type;

	    call_data = beta;

	    if (type != X1f4_E4_CASE) {
		l_USER(&call_data->dxcast, output);

		call_data->dxcast.call = 0;
		call_data->dxcast.type = type;

		output = &call_data->dxcast.data;
	    } else {
	    }

	    call = head;
	    do {
		if (call == count) {
		    call = 0;
		}

		ever_call
		    (frame, track, parts, flags, track[-1],
		     (int *) function_call->args, instant_data,
		     function_text->args,
		     function_data->args, class, lock, count, call);

		call_data->linetext = *linetext_data;

		call_data->linetext.function.type = X1f4_E4_CASE;

		status = _libx1f4i0_lxcall_post_call
		    (&call_data->linetext, function_call, output, 0, frame,
		     POST_BITS,
		     &instant_data->v, &shuffle_data->p, &shuffle_data->q,
		     &shuffle_data->x, shuffle_data->y.eelookup_data);
		if (status) {
		    break;
		}

		call++;
	    } while (call ^ tail);
	}
    }

    return status;
}


static int
part_call(void *back, int status, void *output,
	  struct x1f4_dispatch_type **dispatch)
{
    if (status) {
    } else {
	struct slip_type *call_data;

	call_data = back;
	status = call_data->call(call_data->back, 0, output, dispatch);
	if (status) {
	} else {
	    int slip;

	    slip = call_data->slip;
	    slip++;
	    if (slip != call_data->tail) {
		struct a1_instant_type *instant_data;
		struct a1_shuffle_type *shuffle_data;
		void **frame, **track;

		if (slip == call_data->miss) {
		    slip = 0;
		}

		call_data->slip = slip;

		instant_data = call_data->instant_data;
		shuffle_data = call_data->shuffle_data;

		track = call_data->track;

		frame = (void *) (call_data + 1);
		frame++;

		ever_call
		    (frame, track, call_data->parts, call_data->flags,
		     track[-1], (int *) call_data->function.args,
		     instant_data, call_data->function_text->args,
		     call_data->function_data->args, call_data->class,
		     call_data->lock, call_data->miss, call_data->slip);

		call_data->linetext = *call_data->linetext_data;

		call_data->linetext.function.type = X1f4_E4_CASE;

		status = _libx1f4i0_lxcall_post_call
		    (&call_data->linetext, &call_data->function,
		     (void *) x1f4_a1_walk_link, 0, call_data->clear,
		     POST_BITS, &instant_data->v,
		     &shuffle_data->p, &shuffle_data->q, &shuffle_data->x,
		     shuffle_data->y.eelookup_data);
		if (status) {
		} else {
		    struct x1f4_dispatch_type *dispatch_data;

		    dispatch_data = call_data->clear[2];

		    if (dispatch_data) {
			call_data->back = dispatch_data->back;
			call_data->call = dispatch_data->call;

			dispatch_data->back = call_data;
			dispatch_data->call = part_call;

			call_data->linetext.function.type =
			    call_data->dxcast.type;

			*dispatch = dispatch_data;
		    }
		}
	    } else {
		*dispatch = NULL;
	    }
	}
    }

    return status;
}
