/*
 * e4.7.1.c
 * Copyright (C) 2006-2017, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#include <e4-config.h>

#if defined HAVE_LIBx1f4i0
# include <libx1f4i0.h>
#endif				/* HAVE_LIBx1f4i0 */
#include <stddef.h>
#include <stdlib.h>
#include <string.h>

#include <ctype.0.h>
#include <e4-byte.h>
#include <e4-defs.h>
#include <e4-inter.h>
#include <e4-parse.h>
#include <e4-types.h>
#include <e4fine-e4.h>

#define SCALAR_BITS \
    ((1 << BILL) | (1 << MODE) | (1 << REAL))

#define EXPRESSION(expression) \
    ((e4_expression_type *) expression)

#define E_FREE(mcontext, mdata) \
    (EXPRESSION(mcontext)->m.free(EXPRESSION(mcontext)->m.data, (mdata)))
#define E_LINK(mcontext, mdata, size) \
    (EXPRESSION(mcontext)->m.link					      \
	(EXPRESSION(mcontext)->m.data, (void *) (mdata), (size)))
#define E_MODE(mcontext, mdata, size) \
    (EXPRESSION(mcontext)->m.mode					      \
	(EXPRESSION(mcontext)->m.data, (void *) (mdata), (size)))

#define M_FREE(mcontext, mdata) \
    (mcontext)->m.free((mcontext)->m.data, (mdata))
#define M_LINK(mcontext, mdata, size) \
    (mcontext)->m.link((mcontext)->m.data, (void *) (mdata), (size))
#define M_MODE(mcontext, mdata, size) \
    (mcontext)->m.mode((mcontext)->m.data, (void *) (mdata), (size))

#if __ALLOW_FUNCTION_POINTERS__
# define case_odb2(lock_data, lock_type, call, parser_data) \
    ((lock_data)->odbx_data == _x1f4_e4_pick				      \
     ? 0 : lead_odb2((lock_data), (lock_type), (call), (parser_data)))
#else
# define case_odb2(lock_data, lock_type, call, parser_data) \
    lead_odb2((lock_data), (lock_type), (call), (parser_data))
#endif				/* __ALLOW_FUNCTION_POINTERS__ */

typedef struct half_type {
    int *none;
    struct e4_hack_type *hack_data;
} half_type;

typedef struct lock_type {
    int type;
    struct e4_atom_type *atom_data;
    const struct e4_odbx_type *odbx_data;
    struct lock_type *lock_data;
} lock_type;

extern const struct e4_odbx_type _x1f4_e4_implicit[], _x1f4_e4_pick[];

static int call_hack(struct e4_hack_type *, struct parser_type *);
static int call_link(struct e4_hack_type *, struct e4_atom_type **,
		     struct e4_odb2_type *, struct e4_odb2_type *,
		     struct e4_odb2_type *, const struct e4_odbx_type *,
		     const struct e4_2nde_type *, unsigned, int *, unsigned,
		     struct parser_type *);
static int call_rail(struct e4_hack_type *, struct e4_atom_type **,
		     struct e4_odb2_type *, struct e4_odb2_type *,
		     struct e4_odb2_type *, const struct e4_odbx_type *,
		     const struct e4_2nde_type *, unsigned, int *, unsigned,
		     struct parser_type *);
static int case_atom(struct e4_atom_type *, int *, struct parser_type *);
static int dash_hack(unsigned, const int *, struct e4_hack_type *);
static int deck_atom(struct e4_atom_type *, struct parser_type *);
static int flat_hack(struct e4_hack_type *);
static int head_list(struct e4_hack_type **, const struct e4_line_type *,
		     struct parser_type *);
static int keep_list(struct e4_hack_type **, const struct e4_line_type *,
		     struct e4_atom_type *, unsigned, unsigned, unsigned,
		     struct parser_type *);
static int land_list(struct e4_hack_type **, const struct e4_line_type *,
		     struct e4_atom_type *, unsigned, unsigned,
		     struct parser_type *);
static int lead_ever(struct e4_atom_type *, unsigned);
static int lead_hack(struct e4_hack_type *);
static int lead_odb2(struct lock_type *, int *, int *, struct parser_type *);
static int lead_side(struct e4_atom_type *);
static int lead_type(struct e4_hack_type *, int, int, struct parser_type *,
		     const struct e4_odbx_type *);
static int line_hack(struct e4_hack_type *, const int *, unsigned, void *);
static int link_hack(struct e4_hack_type *, int *, void *);
static int mind_ever(struct e4_atom_type *, struct half_type *,
		     struct parser_type *);
static int mind_hack(struct e4_hack_type *, int *, void *);
static int mind_this(struct e4_atom_type *, struct half_type *,
		     struct parser_type *);
static int near_bill(struct e4_atom_type *, struct half_type *,
		     struct parser_type *);
static int near_ever(struct e4_atom_type *, struct half_type *,
		     struct parser_type *);
static int near_hack(struct e4_hack_type *, int *, void *);
static int near_last(struct e4_atom_type *, struct half_type *,
		     struct parser_type *);
static int near_mode(struct e4_atom_type *, struct half_type *,
		     struct parser_type *);
static int near_real(struct e4_atom_type *, struct half_type *,
		     struct parser_type *);
static int near_text(struct e4_atom_type *, struct half_type *,
		     struct parser_type *);
static int near_this(struct e4_atom_type *, struct half_type *,
		     struct parser_type *);
static int pass_type(struct e4_odb1_type *, int, int);
static int pick_lock(struct lock_type **, struct lock_type **,
		     struct parser_type *);
static int push_lock(struct lock_type **, struct e4_atom_type *, int,
		     const struct e4_odbx_type *, struct parser_type *);
static int type_call(struct e4_hack_type *, int *, int, int,
		     struct parser_type *);
static int type_hack(struct e4_hack_type *, int *, int, int,
		     struct parser_type *);
static int type_last(struct e4_hack_type *, int *, const int *, unsigned,
		     const struct e4_last_type *, struct e4_atom_type *,
		     struct parser_type *);
static int type_miss(struct e4_hack_type *, int *, const int *, unsigned,
		     const struct e4_last_type *, struct e4_atom_type *,
		     struct parser_type *);
static int type_post(struct e4_hack_type *, unsigned,
		     const struct e4_last_type *, struct e4_atom_type *,
		     struct parser_type *);
static int type_trip(int *, struct e4_atom_type *, unsigned,
		     struct parser_type *);

static void free_odb1(struct e4_odb1_type *, void *);
static void free_odbx(struct e4_hack_type *, void *);
#if __ALLOW_RIGHT_FIRST__
static void swap_call(struct e4_atom_type *, unsigned);
#endif				/* __ALLOW_RIGHT_FIRST__ */

static int (*const mind_atom[])(struct e4_atom_type *, struct half_type *,
				struct parser_type *) = {
/* *INDENT-OFF* */
    near_bill,
    near_mode,
    near_real,
    near_text,
    mind_ever,
    near_last,
    near_last,
    near_last,
    mind_this
/* *INDENT-ON* */
}, (*const near_atom[])(struct e4_atom_type *, struct half_type *,
			struct parser_type *) = {
/* *INDENT-OFF* */
    near_bill,
    near_mode,
    near_real,
    near_text,
    near_ever,
    near_last,
    near_last,
    near_last,
    near_this
/* *INDENT-ON* */
};

static int
call_hack(struct e4_hack_type *hack_data, struct parser_type *parser_data)
{
    int status = 0;
    struct e4_atom_type *atom_data;
    unsigned count;

    atom_data = hack_data->atoms;
    count = hack_data->count;
    do {
	struct e4_odb2_type *odb2_data;

	odb2_data = atom_data->odb2._2nd;
	if (odb2_data) {
	    struct e4_odb2_type odb2, *odb2_link = NULL;
	    unsigned land = 0;

	    do {
		int reset = 0;
		const struct e4_odbx_type *odbx_data;

		odbx_data = odb2_data->odbx_data;

#if __ALLOW_RIGHT_FIRST__
		if (odbx_data->bits & LEFT_LAST) {
		    struct e4_atom_type *lead_atom, *miss_atom;
		    unsigned left, post;

		    miss_atom = atom_data;
		    while (land) {
			struct e4_odb2_type *odb2_text;

			miss_atom--;
			land--;
			odb2_text = miss_atom->odb2._2nd;
			while (odb2_text) {
			    odb2_text = odb2_text->odb2_data;
			    land++;
			}
		    }
		    lead_atom = miss_atom;
		    land = 1;
		    do {
			struct e4_odb2_type *odb2_text;

			lead_atom--;
			land--;
			odb2_text = lead_atom->odb2._2nd;
			while (odb2_text) {
			    odb2_text = odb2_text->odb2_data;
			    land++;
			}
		    } while (land);

		    left = miss_atom - lead_atom;
		    post = atom_data - miss_atom + 1;

		    swap_call(lead_atom, left + post);

		    miss_atom = lead_atom + post;

		    swap_call(lead_atom, post);
		    swap_call(miss_atom, left);

		    miss_atom--;

		    if (odb2_link) {
			odb2_link->odb2_data = NULL;
		    } else {
			miss_atom->odb2._2nd = NULL;
		    }

		    odb2_link = atom_data->odb2._2nd;
		    if (odb2_link) {
			struct e4_odb2_type *odb2_call;

			while (1) {
			    land++;
			    odb2_call = odb2_link->odb2_data;
			    if (odb2_call) {
				land++;
				odb2_link = odb2_call->odb2_data;
				if (odb2_link) {
				} else {
				    odb2_call->odb2_data = odb2_data;
				    odb2_link = odb2_call;
				    if (1) {
					break;
				    }
				}
			    } else {
				odb2_link->odb2_data = odb2_data;
				if (1) {
				    break;
				}
			    }
			}
		    } else {
			atom_data->odb2._2nd = odb2_data;
		    }
		}
#endif				/* __ALLOW_RIGHT_FIRST__ */

		if (odbx_data->bits & E2ND_LINK) {
		    const struct e4_2nde_type *e2nd_data;

		    e2nd_data = odbx_data->extension2;
		    if (e2nd_data->bits & (BACK_RAIL | CALL_LINK)) {
			if (e2nd_data->bits & BACK_RAIL) {
			    status = call_rail
				(hack_data, &atom_data, odb2_data, odb2_link,
				 &odb2, odbx_data, e2nd_data, land, &reset,
				 count, parser_data);
			    if (status) {
				break;
			    } else {
				land = -1;

				odb2_data = &odb2;
			    }
			} else {
			    status = call_link
				(hack_data, &atom_data, odb2_data, odb2_link,
				 &odb2, odbx_data, e2nd_data, land, &reset,
				 count, parser_data);
			    if (status) {
				break;
			    } else {
				land = -1;

				odb2_data = &odb2;
			    }
			}
		    }
		}

		land++;

		if (reset) {
		    odb2_link = NULL;
		} else {
		    odb2_link = odb2_data;
		}

		odb2_data = odb2_data->odb2_data;
	    } while (odb2_data);
	    if (status) {
		break;
	    }
	}

	count--;
	atom_data++;
    } while (count);

    if (status) {
    } else {
	if (parser_data->bits & PACK_ESTORAGE) {
	    M_MODE(parser_data, &hack_data->atoms,
		   hack_data->count * sizeof(struct e4_atom_type));
	}

	atom_data = hack_data->atoms;
	count = hack_data->count;
	do {
	    if (atom_data->type == EVER) {
		((struct e4_ever_type *) atom_data->data.ever.ever_data)->bits
		    &= ~SUSPENDED;
	    }

	    count--;
	    atom_data++;
	} while (count);
    }

    return status;
}


static int
call_link(struct e4_hack_type *hack_data, struct e4_atom_type **atom,
	  struct e4_odb2_type *odb2_data, struct e4_odb2_type *odb2_link,
	  struct e4_odb2_type *odb2_trap, const struct e4_odbx_type *odbx_data,
	  const struct e4_2nde_type *e2nd_data, unsigned land, int *reset,
	  unsigned count, struct parser_type *parser_data)
{
    int status;
    struct e4_atom_type *atom_data, *lead_atom, *miss_atom;
    struct e4_hack_type *hack_text;
    const struct e4_line_type *line_data;
    unsigned miss;

    atom_data = *atom;

    miss_atom = atom_data;
    while (land) {
	struct e4_odb2_type *odb2_text;

	miss_atom--;
	land--;
	odb2_text = miss_atom->odb2._2nd;
	while (odb2_text) {
	    odb2_text = odb2_text->odb2_data;
	    land++;
	}
    }

    lead_atom = miss_atom;
    land = 1;
    do {
	struct e4_odb2_type *odb2_text;

	lead_atom--;
	land--;
	odb2_text = lead_atom->odb2._2nd;
	while (odb2_text) {
	    odb2_text = odb2_text->odb2_data;
	    land++;
	}
    } while (land);

    line_data = e2nd_data->line;

    miss = atom_data - miss_atom;

    status = land_list
	(&hack_text, line_data, lead_atom, miss_atom - lead_atom, miss + 1,
	 parser_data);
    if (status) {
    } else {
	const int *args;
	struct e4_odb2_type *odb2_lock;
	unsigned class;

	odb2_lock = odb2_data->odb2_data;

	*reset = 1;

	args = line_data->last.args + 2;

	class = line_data->last.bits;

	line_hack(hack_text, args, class, parser_data->dddevice);
	dash_hack(class, args, hack_text);

	lead_atom->odb2._2nd = odb2_lock;
	lead_atom->data.last.base = hack_text;

	lead_atom->odb1 = NULL;

	if (class & SLIP_LIST) {
	    lead_atom->load = _x1f4_e4_load_lock;
	    lead_atom->type = LOCK;
	} else {
	    lead_atom->type = LAST;
	    if (class & TEXT_LINK) {
		_x1f4_e4_link_laSt(lead_atom);
	    } else {
		_x1f4_e4_link_last(lead_atom);
	    }
	}

	hack_text++;
	args++;

	if (odb2_link) {
	    odb2_link->odb2_data = NULL;
	} else {
	    ((struct e4_atom_type *) hack_text->atoms)[0].odb2._2nd = NULL;
	}

	line_hack(hack_text, args, class, parser_data->dddevice);
	dash_hack(class, args, hack_text);

	if (count ^ 1) {
	    memmove
		(lead_atom + 1, atom_data + 1,
		 (count - 1) * sizeof(struct e4_atom_type));
	}

	hack_data->count -= atom_data - lead_atom;

	*atom = lead_atom;

	M_FREE(parser_data, odb2_data);

	odb2_trap->odb2_data = odb2_lock;
    }

    return status;
}


static int
call_rail(struct e4_hack_type *hack_data, struct e4_atom_type **atom,
	  struct e4_odb2_type *odb2_data, struct e4_odb2_type *odb2_link,
	  struct e4_odb2_type *odb2_trap, const struct e4_odbx_type *odbx_data,
	  const struct e4_2nde_type *e2nd_data, unsigned land, int *reset,
	  unsigned count, struct parser_type *parser_data)
{
    int status;
    struct e4_atom_type *atom_data, *kick_atom, *lead_atom, *miss_atom;
    struct e4_hack_type *hack_text;
    const struct e4_line_type *line_data;
    struct e4_odb2_type *odb2_rail = NULL;
    unsigned miss;

    atom_data = *atom;

    miss_atom = atom_data;
    while (land) {
	struct e4_odb2_type *odb2_text;

	miss_atom--;
	land--;
	odb2_text = miss_atom->odb2._2nd;
	while (odb2_text) {
	    odb2_text = odb2_text->odb2_data;
	    land++;
	}
    }

    /*
     * is there ever more than one hack_data atom in the mid section?
     */
    lead_atom = miss_atom;
    land = -1;
    if (1) {
	struct e4_odb2_type *odb2_text;

	lead_atom--;
	odb2_text = lead_atom->odb2._2nd;
	do {
	    land++;
	    odb2_rail = odb2_text;
	    odb2_text = odb2_text->odb2_data;
	    /* odb2_text always nil */
	} while (odb2_text);
    }
    /* land always 0 */
    while (land) {
	struct e4_odb2_type *odb2_text;

	lead_atom--;
	land--;
	odb2_text = lead_atom->odb2._2nd;
	while (odb2_text) {
	    odb2_text = odb2_text->odb2_data;
	    land++;
	}
    }

    kick_atom = lead_atom;
    land = 1;
    do {
	struct e4_odb2_type *odb2_text;

	kick_atom--;
	land--;
	odb2_text = kick_atom->odb2._2nd;
	while (odb2_text) {
	    land++;

	    odb2_text = odb2_text->odb2_data;
	}
    } while (land);

    line_data = e2nd_data->line;

    miss = atom_data - miss_atom;

    status = keep_list
	(&hack_text, line_data, kick_atom, lead_atom - kick_atom,
	 miss_atom - lead_atom, miss + 1, parser_data);
    if (status) {
    } else {
	const int *args;
	struct e4_odb2_type *odb2_lock;
	unsigned class;

	odb2_lock = odb2_data->odb2_data;

	*reset = 1;

	args = line_data->last.args + 3;

	class = line_data->last.bits;

	line_hack(hack_text, args, class, parser_data->dddevice);
	dash_hack(class, args, hack_text);

	kick_atom->odb2._2nd = odb2_lock;
	kick_atom->data.last.base = hack_text;

	kick_atom->odb1 = NULL;

	if (class & SLIP_LIST) {
	    kick_atom->load = _x1f4_e4_load_lock;
	    kick_atom->type = LOCK;
	} else {
	    kick_atom->type = LAST;
	    if (class & TEXT_LINK) {
		_x1f4_e4_link_laSt(kick_atom);
	    } else {
		_x1f4_e4_link_last(kick_atom);
	    }
	}

	hack_text++;
	args++;

	line_hack(hack_text, args, class, parser_data->dddevice);
	dash_hack(class, args, hack_text);

	hack_text++;
	args++;

	if (odb2_link) {
	    odb2_link->odb2_data = NULL;
	} else {
	    ((struct e4_atom_type *) hack_text->atoms)[0].odb2._2nd = NULL;
	}

	line_hack(hack_text, args, class, parser_data->dddevice);
	dash_hack(class, args, hack_text);

	if (count ^ 1) {
	    memmove
		(kick_atom + 1, atom_data + 1,
		 (count - 1) * sizeof(struct e4_atom_type));
	}

	hack_data->count -= atom_data - kick_atom;

	*atom = kick_atom;

	M_FREE(parser_data, odb2_data);

	M_FREE(parser_data, odb2_rail);

	odb2_trap->odb2_data = odb2_lock;
    }

    return status;
}


static int
case_atom(struct e4_atom_type *atom_data, int *atom_type,
	  struct parser_type *parser_data)
{
    int status = 0;
    const struct e4_odbx_type *const *list;

    list = parser_data->attributes_data->odb1_data;
    if (!list) {
	status = _x1f4_e4_080704_error_g(parser_data);
    } else {
	int type;
	struct e4_odb1_type *base_odb1 = NULL, *last_odb1 = NULL, *odb1_data;

	type = *atom_type;

	odb1_data = atom_data->odb1;
	while (odb1_data) {
	    const char *e;
	    struct text_type *text_data;
	    unsigned length;

	    text_data = (struct text_type *) (odb1_data + 1);

	    e = text_data->data;
	    length = text_data->size;
	    while (length) {
		const struct e4_odbx_type *const *odbx, *odbx_data,
		    *odbx_text = NULL;
		unsigned toxic = 0;

		odbx = list;

		odbx_data = *odbx;
		while (odbx_data) {
		    unsigned shift;

		    shift = odbx_data->size;
		    if (toxic < shift) {
			if (shift < length + 1) {
			    if (*odbx_data->args == type) {
				const char *name;

				name = odbx_data->name;
				if (memcmp
				    ((void *) (e - shift), name, shift)) {
				} else {
				    odbx_text = odbx_data;
				    toxic = shift;
				}
			    }
			}
		    }

		    odbx++;
		    odbx_data = *odbx;
		}
		if (odbx_text) {
		    struct e4_odb1_type *ocb1_data;

		    e -= toxic;
		    length -= toxic;

		    status = M_LINK(parser_data, &ocb1_data,
				    sizeof(struct e4_odb1_type));
		    if (status) {
			status = ALLOC_ERROR;
			break;
		    } else {
			if (base_odb1) {
			    last_odb1->odb1_data = ocb1_data;
			} else {
			    base_odb1 = ocb1_data;
			}

			last_odb1 = ocb1_data;

			ocb1_data->odb1_data = NULL;
			ocb1_data->odbx_data = odbx_text;

			type = odbx_text->type;
		    }
		} else {
		    status = _x1f4_e4_080704_error_i(parser_data, e, length);
		    break;
		}
	    }
	    if (status) {
		break;
	    }

	    odb1_data = odb1_data->odb1_data;
	}

	if (status) {
	    free_odb1(base_odb1, parser_data->expression_data);
	} else {
	    *atom_type = type;

	    free_odb1(atom_data->odb1, parser_data->expression_data);

	    atom_data->odb1 = base_odb1;
	}
    }

    return status;
}


static int
dash_hack(unsigned class, const int *args, struct e4_hack_type *hack_data)
{
    struct e4_atom_type *atom_data;
    unsigned count;

    atom_data = hack_data->atoms;
    count = hack_data->count;
    if (count == 1) {
	if (class & POST_TYPE
	    && (*args & (POST_FLAT | POST_XSET)) == (POST_FLAT | POST_XSET)) {
	    int type;

	    type = atom_data->type;

	    if (type == THIS) {
		struct e4_hack_type *hack_data;

		while (type == THIS) {
		    hack_data = atom_data->data.hack.data;
		    if (hack_data->count == 1) {
			atom_data = hack_data->atoms;
			type = atom_data->type;
		    } else {
			break;
		    }
		}
		if (hack_data->bits & ZERO_LONG) {
		    if (type == EVER) {
			((struct e4_ever_type *)
			 atom_data->data.ever.ever_data)->bits &= ~PRINCIPAL;
		    }
		}
	    } else {
		if (type == EVER) {
		    struct e4_ever_type *ever_data;
		    unsigned bits;

		    ever_data = (void *) atom_data->data.ever.ever_data;
		    bits = ever_data->bits;
		    if (bits & SUSPENDED) {
			ever_data->bits = bits & ~(PRINCIPAL | SUSPENDED);
		    }
		}
	    }
	} else {
	    if (atom_data->type == EVER) {
		((struct e4_ever_type *) atom_data->data.ever.ever_data)->bits
		    &= ~SUSPENDED;
	    }
	}
    } else {
	do {
	    if (atom_data->type == EVER) {
		((struct e4_ever_type *) atom_data->data.ever.ever_data)->bits
		    &= ~SUSPENDED;
	    }

	    count--;
	    atom_data++;
	} while (count);
    }

    return 0;
}


static int
deck_atom(struct e4_atom_type *lead_atom, struct parser_type *parser_data)
{
    struct e4_odb1_type *odb1_data, *odb1_text = NULL;
    unsigned status = 0;

    odb1_data = lead_atom->odb1;
    while (odb1_data) {
	struct e4_odb1_type *odb1_lock;
	const struct e4_odbx_type *odbx_data;

	odb1_lock = odb1_data->odb1_data;

	odbx_data = odb1_data->odbx_data;
	if (odbx_data->bits & E2ND_LINK) {
	    const struct e4_2nde_type *e2nd_data;

	    e2nd_data = odbx_data->extension2;
	    if (e2nd_data->bits & CALL_LINK) {
		struct e4_hack_type *hack_text = NULL;
		const struct e4_line_type *line_data;

		line_data = e2nd_data->line;

		status = head_list(&hack_text, line_data, parser_data);
		if (status) {
		    break;
		} else {
		    struct e4_atom_type *atom_data;
		    unsigned class;

		    class = line_data->last.bits;

		    atom_data = hack_text->atoms;

		    *atom_data = *lead_atom;

		    atom_data->odb2._2nd = NULL;

		    if (odb1_text) {
			odb1_text->odb1_data = NULL;
		    } else {
			atom_data->odb1 = NULL;
		    }

		    line_hack(hack_text, NULL, class, parser_data->dddevice);

		    lead_atom->type = LAST;
		    lead_atom->data.last.base = hack_text;
		    lead_atom->odb1 = odb1_lock;

		    if (class & TEXT_LINK) {
			_x1f4_e4_link_laSt(lead_atom);
		    } else {
			_x1f4_e4_link_last(lead_atom);
		    }

		    M_FREE(parser_data, odb1_data);
		}
	    } else {
		odb1_text = odb1_data;
	    }
	} else {
	    odb1_text = odb1_data;
	}

	odb1_data = odb1_lock;
    }

    return status;
}


static int
flat_hack(struct e4_hack_type *hack_data)
{
    struct e4_atom_type *atom_data;

    atom_data = hack_data->atoms;
    if (atom_data->type == EVER) {
	((struct e4_ever_type *) atom_data->data.ever.ever_data)->bits
	    &= ~SUSPENDED;
    }

    return 0;
}


static int
head_list(struct e4_hack_type **hack, const struct e4_line_type *line_data,
	  struct parser_type *parser_data)
{
#define BACK_HACK(count, excess) \
    (offset + sizeof(struct e4_hack_type))

    int status;
    struct e4_hack_type *hack_data;
    unsigned offset;

    offset = sizeof(struct e4_base_type);

    status = M_LINK(parser_data, &hack_data, BACK_HACK(count, excess));
    if (status) {
	status = ALLOC_ERROR;
    } else {
	void *fast;

	status = M_LINK(parser_data, &fast, sizeof(struct e4_atom_type));
	if (status) {
	    status = ALLOC_ERROR;

	    M_FREE(parser_data, hack_data);
	} else {
	    hack_data = (struct e4_hack_type *) ((char *) hack_data + offset);

	    hack_base(hack_data)->last_data = &line_data->last;

	    if (line_data->last.bits & TEXT_LINK) {
		hack_base(hack_data)->text = (void *) line_data->text;
	    }

	    *hack = hack_data;

	    hack_data->count = 1;
	    hack_data->atoms = fast;
	    /*
	     * likely unused, will set to a possibly incorrect value
	     */
	    hack_data->bits = READ_ONLY;
	}
    }

    return status;

#undef BACK_HACK
}


static int
keep_list(struct e4_hack_type **hack, const struct e4_line_type *line_data,
	  struct e4_atom_type *atom_data, unsigned kick, unsigned lead,
	  unsigned miss, struct parser_type *parser_data)
{
#define BLUE_HACK(count) \
    (sizeof(struct e4_hack_type) * 3)
#define BACK_HACK(count, excess) \
    (offset + BLUE_HACK(count) + excess)

    int status;
    struct e4_hack_type *hack_data = NULL, *free_hack = NULL;
    void *dash = NULL, *ever = NULL, *fast = NULL;

    /*
     * midsection is one hack atom only, might as well optimize for that
     */

    do {
	unsigned excess = 0, offset;

	offset = sizeof(struct e4_base_type);
	if (line_data->last.bits & SLIP_LIST) {
	    /*
	     * for time being, recognize no SLIP LIST operators.
	     */
	    status = PARSE_ERROR;
	    break;

	    offset += sizeof(struct e4_fine_type);
	    excess = sizeof(struct e4_ever_type) + sizeof(void *);
	    if (parser_data->e4fine) {
	    } else {
		status = PARSE_ERROR;
		break;
	    }
	}

	status = M_LINK(parser_data, &hack_data, BACK_HACK(count, excess));
	if (status) {
	    status = ALLOC_ERROR;
	    break;
	}

	if (1) {
	    status =
		M_LINK(parser_data, &fast, kick * sizeof(struct e4_atom_type));
	    if (status) {
		status = ALLOC_ERROR;
		break;
	    }
	}

	if (lead != 1 || atom_data[kick].type != THIS) {
	    status =
		M_LINK(parser_data, &ever, lead * sizeof(struct e4_atom_type));
	    if (status) {
		status = ALLOC_ERROR;
		break;
	    }
	}

	if (1) {
	    status =
		M_LINK(parser_data, &dash, miss * sizeof(struct e4_atom_type));
	    if (status) {
		status = ALLOC_ERROR;
		break;
	    }
	}

	hack_data = (struct e4_hack_type *) ((char *) hack_data + offset);

	hack_base(hack_data)->last_data = &line_data->last;

	if (line_data->last.bits & TEXT_LINK) {
	    hack_base(hack_data)->text = (void *) line_data->text;
	}

	*hack = hack_data;

	hack_data->count = kick;
	hack_data->atoms = fast;
	/*
	 * likely unused, will set to a possibly incorrect value
	 */
	hack_data->bits = READ_ONLY;

	memcpy(fast, atom_data, kick * sizeof(struct e4_atom_type));

	hack_data++;

	atom_data += kick;

	if (ever) {
	    struct e4_atom_type *ever_atom;

	    hack_data->count = lead;
	    hack_data->atoms = ever;
	    /*
	     * likely unused, will set to a possibly incorrect value
	     */
	    hack_data->bits = READ_ONLY & 0;

	    memcpy(ever, atom_data, lead * sizeof(struct e4_atom_type));

	    ever_atom = ever;
	    if (lead ^ 1) {
		struct e4_odb2_type *odb2_data, *odb2_slip, *odb2_tail;

		ever_atom += lead - 1;
		odb2_data = ever_atom->odb2._2nd;
		odb2_slip = odb2_data->odb2_data;
		odb2_tail = odb2_slip->odb2_data;
		while (odb2_tail) {
		    odb2_data = odb2_slip;
		    odb2_slip = odb2_tail;
		    odb2_tail = odb2_slip->odb2_data;
		}

		odb2_data->odb2_data = NULL;
	    } else {
		ever_atom->odb2._2nd = NULL;
	    }
	} else {
	    free_hack = atom_data->data.hack.data;

	    *hack_data = *free_hack;
	}

	hack_data++;

	hack_data->count = miss;
	hack_data->atoms = dash;
	/*
	 * likely unused, will set to a possibly incorrect value
	 */
	hack_data->bits = READ_ONLY & 0;

	memcpy(dash, atom_data + lead, miss * sizeof(struct e4_atom_type));

	if (excess) {
	    const int *clip;
	    int args[6], type;
	    struct e4_line_type land;
	    const struct e4_line_type **line;

	    hack_data -= 2;

	    clip = line_data->last.args;

	    args[0] = clip[0];
	    args[1] = clip[1];
	    args[2] = clip[2];
	    if (line_data->last.bits & POST_TYPE) {
		args[3] = clip[3];
		args[4] = clip[4];
		args[5] = clip[5];
	    }

	    type_trip(args + 0, atom_data - kick, kick, parser_data);
	    type_trip(args + 1, atom_data, lead, parser_data);
	    type_trip(args + 2, atom_data + lead, miss, parser_data);

	    type = args[0];

	    land.last = line_data->last;
	    land.last.args = args;
	    land.last.type = type;

	    status = x1f4_type_e4fine(parser_data->e4fine, &type, &land);
	    if (status) {
		status = ALLOC_ERROR;
	    } else {
		struct e4_ever_type *ever_data;
		struct e4_fine_type *fine_data;
		const void *call;

		ever_data = (void *)
		    ((char *) hack_data + BLUE_HACK(count));
		line = (void *) (ever_data + 1);

		fine_data = hack_fine(hack_data);

		ever_data->bits = 0;
		ever_data->name = line_data->last.name;
		ever_data->size = line_data->last.size;
		ever_data->type = type;

		*line = line_data;

		fine_data->ever_data = ever_data;
		fine_data->line = line;

		x1f4_pick_e4fine(parser_data->e4fine, type, &call);
		hack_base(hack_data)->last_data =
		    &((struct e4_line_type *) call)->last;
	    }
	}
    } while (0);

    if (status) {
	if (ever) {
	    M_FREE(parser_data, dash);
	}
	if (ever) {
	    M_FREE(parser_data, ever);
	}
	if (fast) {
	    M_FREE(parser_data, fast);
	}
	if (hack_data) {
	    M_FREE(parser_data, hack_data);
	}
    } else {
	if (free_hack) {
	    M_FREE(parser_data, free_hack);
	}
    }

    return status;

#undef BACK_HACK
#undef BLUE_HACK
}


static int
land_list(struct e4_hack_type **hack, const struct e4_line_type *line_data,
	  struct e4_atom_type *atom_data, unsigned lead, unsigned miss,
	  struct parser_type *parser_data)
{
#define BLUE_HACK(count) \
    (sizeof(struct e4_hack_type) * 2)
#define BACK_HACK(count, excess) \
    (offset + BLUE_HACK(count) + excess)

    int status;
    struct e4_hack_type *hack_data = NULL;
    void *ever = NULL, *fast = NULL;

    do {
	unsigned excess = 0, offset;

	offset = sizeof(struct e4_base_type);
	if (line_data->last.bits & SLIP_LIST) {
	    offset += sizeof(struct e4_fine_type);
	    excess = sizeof(struct e4_ever_type) + sizeof(void *);
	    if (parser_data->e4fine) {
	    } else {
		status = PARSE_ERROR;
		break;
	    }
	}

	status = M_LINK(parser_data, &hack_data, BACK_HACK(count, excess));
	if (status) {
	    status = ALLOC_ERROR;
	    break;
	}

	status =
	    M_LINK(parser_data, &fast, lead * sizeof(struct e4_atom_type));
	if (status) {
	    status = ALLOC_ERROR;
	    break;
	}

	status =
	    M_LINK(parser_data, &ever, miss * sizeof(struct e4_atom_type));
	if (status) {
	    status = ALLOC_ERROR;
	    break;
	}

	hack_data = (struct e4_hack_type *)
	    ((char *) hack_data + offset);

	hack_base(hack_data)->last_data = &line_data->last;

	if (line_data->last.bits & TEXT_LINK) {
	    hack_base(hack_data)->text = (void *) line_data->text;
	}

	*hack = hack_data;

	hack_data->count = lead;
	hack_data->atoms = fast;
	/*
	 * likely unused, will set to a possibly incorrect value
	 */
	hack_data->bits = READ_ONLY;

	memcpy(fast, atom_data, lead * sizeof(struct e4_atom_type));

	hack_data++;

	hack_data->count = miss;
	hack_data->atoms = ever;
	/*
	 * likely unused, will set to a possibly incorrect value
	 */
	hack_data->bits = READ_ONLY & 0;

	memcpy(ever, atom_data + lead, miss * sizeof(struct e4_atom_type));

	if (excess) {
	    const int *clip;
	    int args[4], type;
	    struct e4_line_type land;
	    const struct e4_line_type **line;

	    hack_data--;

	    clip = line_data->last.args;

	    args[0] = clip[0];
	    args[1] = clip[1];
	    if (line_data->last.bits & POST_TYPE) {
		args[2] = clip[2];
		args[3] = clip[3];
	    }

	    type_trip(args + 0, atom_data, lead, parser_data);
	    type_trip(args + 1, atom_data + lead, miss, parser_data);

	    type = args[0];

	    land.last = line_data->last;
	    land.last.args = args;
	    land.last.type = type;

	    status = x1f4_type_e4fine(parser_data->e4fine, &type, &land);
	    if (status) {
		status = ALLOC_ERROR;
	    } else {
		struct e4_ever_type *ever_data;
		struct e4_fine_type *fine_data;
		const void *call;

		ever_data = (void *)
		    ((char *) hack_data + BLUE_HACK(count));
		line = (void *) (ever_data + 1);

		fine_data = hack_fine(hack_data);

		ever_data->bits = 0;
		ever_data->name = line_data->last.name;
		ever_data->size = line_data->last.size;
		ever_data->type = type;

		*line = line_data;

		fine_data->ever_data = ever_data;
		fine_data->line = line;

		x1f4_pick_e4fine(parser_data->e4fine, type, &call);
		hack_base(hack_data)->last_data =
		    &((struct e4_line_type *) call)->last;
	    }
	}
    } while (0);

    if (status) {
	if (ever) {
	    M_FREE(parser_data, ever);
	}
	if (fast) {
	    M_FREE(parser_data, fast);
	}
	if (hack_data) {
	    M_FREE(parser_data, hack_data);
	}
    }

    return status;

#undef BACK_HACK
#undef BLUE_HACK
}


static int
lead_ever(struct e4_atom_type *atom_data, unsigned bits)
{
    if (bits & UNIVERSAL) {
	if (bits & REFERENCE) {
	    atom_data->load = _x1f4_e4_load_line;
	} else {
	    atom_data->load = _x1f4_e4_load_slip;
	}
    } else {
	if (bits & REFERENCE) {
	    atom_data->load = _x1f4_e4_load_clip;
	} else {
	    atom_data->load = _x1f4_e4_load_ever;
	}
    }

    return 0;
}


static int
lead_hack(struct e4_hack_type *hack_data)
{
    struct e4_atom_type *atom_data;

    while (1) {
	atom_data = hack_data->atoms;
	if (atom_data->type == THIS) {
	    hack_data = atom_data->data.hack.data;
	} else {
	    lead_side(atom_data);
	    break;
	}
    }

    return 0;
}


static int
lead_odb2(struct lock_type *lock_data, int *lock_type, int *call,
	  struct parser_type *parser_data)
{
    int major, minor, status = 0;

    major = lock_data->type;
    minor = *lock_type;

    do {
	int miss_type;
	const int *args;
	const struct e4_odbx_type *odbx_data;
	unsigned bits;

	odbx_data = lock_data->odbx_data;
	args = odbx_data->args;
	if (major != args[0] || minor != args[1]) {
	    status = _x1f4_e4_type_risk
		(&odbx_data, &miss_type, major, minor, odbx_data, parser_data);
	    if (status) {
		break;
	    } else {
		lock_data->odbx_data = odbx_data;
	    }
	} else {
	    miss_type = odbx_data->type;
	}

	bits = odbx_data->bits;

	if (bits & E2ND_LINK) {
	    if (((struct e4_2nde_type *)
		 odbx_data->extension2)->bits & (BACK_RAIL | CALL_LINK)) {
		*call = 1;
	    }
	}

#if __ALLOW_RIGHT_FIRST__
	if (bits & LEFT_LAST) {
	    *call = 1;
	}
#endif				/* __ALLOW_RIGHT_FIRST__ */

	do {
	    if (!(bits & LEFT_XSET)) {
		struct e4_atom_type *atom_data;

		atom_data = lock_data->atom_data;
		if (atom_data) {
		    int type;

		    type = atom_data->type;
		    if (type == EVER) {
			if (atom_data->data.ever.ever_data->bits
			    & READ_ONLY) {
			} else {
			    lead_side(atom_data);
			}
		    } else {
			if (atom_data->type != THIS) {
			} else {
			    struct e4_hack_type *hack_data;

			    hack_data = atom_data->data.hack.data;
			    if (hack_data->bits & READ_ONLY) {
			    } else {
				lead_hack(hack_data);
			    }
			}
		    }
		}
	    } else {
		struct e4_atom_type *atom_data;

		atom_data = lock_data->atom_data;
		if (!atom_data) {
		    status = _x1f4_e4_080704_error_b(parser_data);
		    break;
		} else {
		    int type;

		    type = atom_data->type;
		    if (type == EVER) {
			if (atom_data->data.ever.ever_data->bits
			    & READ_ONLY) {
			    status = _x1f4_e4_080704_error_7
				(parser_data, atom_data);
			    break;
			}
		    } else {
			if (atom_data->type != THIS) {
			    status = _x1f4_e4_080704_error_b(parser_data);
			    break;
			} else {
			    struct e4_hack_type *hack_data;

			    hack_data = atom_data->data.hack.data;
			    if (hack_data->bits & READ_ONLY) {
				status = _x1f4_e4_080704_error_c
				    (parser_data, hack_data);
				break;
			    }
			}
		    }
		}
	    }
	} while (0);
	if (status) {
	    break;
	}

	*lock_type = miss_type;
    } while (0);

    return status;
}


static int
lead_side(struct e4_atom_type *atom_data)
{
    const struct e4_ever_type *ever_data;
    int type;

    ever_data = atom_data->data.ever.ever_data;

    type = ever_data->type;
    if (ever_data->bits & UNIVERSAL) {
	if (ever_data->bits & REFERENCE) {
	    if (type == BILL) {
		atom_data->load = _x1f4_e4_load_Will;
	    } else if (type == MODE) {
		atom_data->load = _x1f4_e4_load_Node;
	    } else if (type == REAL) {
		atom_data->load = _x1f4_e4_load_Teal;
	    } else if (type == TEXT) {
		atom_data->load = _x1f4_e4_load_Next;
	    } else {
		atom_data->load = _x1f4_e4_load_Over;
	    }
	} else {
	    if (type == BILL) {
		atom_data->load = _x1f4_e4_load_WILL;
	    } else if (type == MODE) {
		atom_data->load = _x1f4_e4_load_NODE;
	    } else if (type == REAL) {
		atom_data->load = _x1f4_e4_load_TEAL;
	    } else if (type == TEXT) {
		atom_data->load = _x1f4_e4_load_NEXT;
	    } else {
		atom_data->load = _x1f4_e4_load_OVER;
	    }
	}
    } else {
	if (ever_data->bits & REFERENCE) {
	    if (type == BILL) {
		atom_data->load = _x1f4_e4_load_Bill;
	    } else if (type == MODE) {
		atom_data->load = _x1f4_e4_load_Mode;
	    } else if (type == REAL) {
		atom_data->load = _x1f4_e4_load_Real;
	    } else if (type == TEXT) {
		atom_data->load = _x1f4_e4_load_Text;
	    } else {
		atom_data->load = _x1f4_e4_load_User;
	    }
	} else {
	    if (type == BILL) {
		atom_data->load = _x1f4_e4_load_BILL;
	    } else if (type == MODE) {
		atom_data->load = _x1f4_e4_load_MODE;
	    } else if (type == REAL) {
		atom_data->load = _x1f4_e4_load_REAL;
	    } else if (type == TEXT) {
		atom_data->load = _x1f4_e4_load_TEXT;
	    } else {
		atom_data->load = _x1f4_e4_load_USER;
	    }
	}
    }

    return 0;
}


static int
lead_type(struct e4_hack_type *hack_data, int type, int name,
	  struct parser_type *parser_data,
	  const struct e4_odbx_type *odbx_data)
{
    int status;
    struct e4_odb1_type *odb1_data;

    status = M_LINK(parser_data, &odb1_data, sizeof(struct e4_odb1_type));
    if (status) {
	status = ALLOC_ERROR;
    } else {
	odb1_data->odb1_data = NULL;
	if (hack_data->count == 1) {
	    struct e4_atom_type *atom_data;
	    struct e4_odb1_type *atom_odb1;

	    atom_data = hack_data->atoms;
	    atom_odb1 = atom_data->odb1;
	    if (atom_odb1) {
		while (atom_odb1->odb1_data) {
		    atom_odb1 = atom_odb1->odb1_data;
		}
		atom_odb1->odb1_data = odb1_data;
	    } else {
		atom_data->odb1 = odb1_data;
	    }

	    if (odbx_data) {
		odb1_data->odbx_data = odbx_data;
	    } else {
		pass_type(odb1_data, name, type);
	    }
	} else {
	    struct e4_hack_type *hack_this;

	    status =
		M_LINK(parser_data, &hack_this, sizeof(struct e4_hack_type));
	    if (status) {
		M_FREE(parser_data, odb1_data);
		status = ALLOC_ERROR;
	    } else {
		struct e4_atom_type *atom_data;

		status = M_LINK(parser_data, &atom_data,
				sizeof(struct e4_atom_type));
		if (status) {
		    M_FREE(parser_data, hack_this);
		    M_FREE(parser_data, odb1_data);
		    status = ALLOC_ERROR;
		} else {
		    atom_data->data.hack.data = hack_this;
		    atom_data->load = _x1f4_e4_load_this;
		    atom_data->odb1 = odb1_data;
		    atom_data->odb2._2nd = NULL;
		    atom_data->type = THIS;
		    *hack_this = *hack_data;
		    hack_data->count = 1;
		    hack_data->atoms = atom_data;
		    hack_data->bits = READ_ONLY;
		    _x1f4_e4_fast_hack(hack_data, parser_data->dddevice);

		    if (odbx_data) {
			odb1_data->odbx_data = odbx_data;
		    } else {
			pass_type(odb1_data, name, type);
		    }
		}
	    }
	}
    }

    return status;
}


static int
line_hack(struct e4_hack_type *hack_data, const int *fits, unsigned bits,
	  void *dddevice)
{
    unsigned class;

    _x1f4_e4_fast_hack(hack_data, dddevice);

    if (bits & POST_TYPE) {
	class = *fits;
    } else {
	class = 0;
    }

    _x1f4_e4_fast_miss(hack_data, class);

    return 0;
}


static int
link_hack(struct e4_hack_type *hack_data, int *type, void *parser)
{
    int status;

    if (hack_data->count == 1) {
	status = mind_hack(hack_data, type, parser);
    } else {
	status = _x1f4_e4_trip_hack(hack_data, parser);
	if (status) {
	} else {
	    if (hack_data->count == 1) {
		status = mind_hack(hack_data, type, parser);
	    } else {
		status = near_hack(hack_data, type, parser);
	    }
	}
    }

    return status;
}


static int
mind_ever(struct e4_atom_type *atom_data, struct half_type *half_data,
	  struct parser_type *parser_data)
{
    const struct e4_ever_type *ever_data;
    unsigned bits;

    ever_data = atom_data->data.ever.ever_data;
    bits = ever_data->bits;
    *half_data->none = ever_data->type;
    if (bits & READ_ONLY) {
	lead_side(atom_data);
    } else {
	lead_ever(atom_data, bits);
	if (1) {
	    half_data->hack_data->bits &= ~READ_ONLY;
	}
    }

    if (bits & SUSPENDED) {
	parser_data->post_ever = ever_data;
	half_data->hack_data->bits |= ZERO_LONG;
    }

    return 0;
}


static int
mind_hack(struct e4_hack_type *hack_data, int *type, void *parser)
{
    int hack_type = 0, none, status = 0;
    struct e4_atom_type *atom_data;
    struct half_type half;
    struct parser_type *parser_data;

    parser_data = parser;

    half.hack_data = hack_data;
    half.none = &none;

    atom_data = hack_data->atoms;
    if (1) {
	none = atom_data->type;

	status = mind_atom[none](atom_data, &half, parser_data);
	if (status) {
	} else {
	    atom_data->odb2._2nd = NULL;

	    if (atom_data->odb1) {
		status = case_atom(atom_data, &none, parser_data);
		if (status) {
		    atom_data++;
		} else {
		    status = deck_atom(atom_data, parser_data);
		    if (status) {
			atom_data++;
		    } else {
			hack_data->bits |= READ_ONLY;
		    }
		}
	    }

	    if (1) {
		hack_type = none;
	    }
	}
    }

    flat_hack(hack_data);

    if (status) {
	hack_data->count =
	    atom_data - (struct e4_atom_type *) hack_data->atoms;
	free_odbx(hack_data, parser_data->expression_data);
	hack_data->count = 1;
    } else {
	*type = hack_type;

	_x1f4_e4_fast_hack(hack_data, parser_data->dddevice);
    }

    return status;
}


static int
mind_this(struct e4_atom_type *atom_data, struct half_type *half_data,
	  struct parser_type *parser_data)
{
    int status;

    status = link_hack
	(atom_data->data.hack.data, half_data->none, parser_data);
    if (status) {
    } else {
	if (1) {
	    atom_data->load = _x1f4_e4_load_this;
	    if (1) {
		if (((struct e4_hack_type *) atom_data->data.hack.data)->bits
		    & READ_ONLY) {
		} else {
		    half_data->hack_data->bits &= ~READ_ONLY;
		}
	    }
	}
    }

    return status;
}


static int
near_bill(struct e4_atom_type *atom_data, struct half_type *half_data,
	  struct parser_type *parser_data)
{
    atom_data->load = _x1f4_e4_load_bill;

    return 0;
}


static int
near_ever(struct e4_atom_type *atom_data, struct half_type *half_data,
	  struct parser_type *parser_data)
{
    const struct e4_ever_type *ever_data;
    unsigned bits;

    ever_data = atom_data->data.ever.ever_data;
    bits = ever_data->bits;
    *half_data->none = ever_data->type;
    if (bits & READ_ONLY) {
	lead_side(atom_data);
    } else {
	lead_ever(atom_data, bits);
    }

    return 0;
}


static int
near_hack(struct e4_hack_type *hack_data, int *type, void *parser)
{
    int call = 0, hack_type = 0, none, status = 0;
    struct e4_atom_type *atom_data;
    struct half_type half;
    struct lock_type *lock_data = NULL;
    struct parser_type *parser_data;
    unsigned count, i;

    parser_data = parser;

    count = hack_data->count;

    half.hack_data = hack_data;
    half.none = &none;

    atom_data = hack_data->atoms;
    for (i = count; i; i--) {
	const struct e4_odbx_type *odbx_data;

	none = atom_data->type;

	status = near_atom[none](atom_data, &half, parser_data);
	if (status) {
	    break;
	}

	odbx_data = atom_data->odb2._1st;
	atom_data->odb2._2nd = NULL;

	if (atom_data->odb1) {
	    status = case_atom(atom_data, &none, parser_data);
	    if (status) {
		atom_data++;
		break;
	    } else {
		status = deck_atom(atom_data, parser_data);
		if (status) {
		    atom_data++;
		    break;
		} else {
		    hack_data->bits |= READ_ONLY;
		}
	    }
	}

	if (i == 1) {
	    struct e4_odb2_type *odb2_data = NULL;

	    do {
		void *data;

		status = M_LINK
		    (parser_data, &data, sizeof(struct e4_odb2_type));
		if (status) {
		    status = ALLOC_ERROR;
		    break;
		}
		if (odb2_data) {
		    odb2_data->odb2_data = data;
		    odb2_data = data;
		} else {
		    atom_data->odb2._2nd = data;
		    odb2_data = data;
		}

		odb2_data->odb2_data = NULL;

		status = case_odb2(lock_data, &none, &call, parser_data);
		if (status) {
		    break;
		}

		odb2_data->odbx_data = lock_data->odbx_data;

		pick_lock(&lock_data, NULL, parser_data);
	    } while (lock_data);
	    if (status) {
		atom_data++;
		break;
	    }

	    hack_type = none;
	} else {
	    if (!lock_data) {
		status = push_lock
		    (&lock_data, atom_data, none, odbx_data, parser_data);
		if (status) {
		    atom_data++;
		    break;
		}
	    } else {
		int form, keep, sink;
		const struct e4_odbx_type *odbx_text;

		odbx_text = lock_data->odbx_data;
		form = odbx_data->pass;
		sink = odbx_text->pass;

		if (form < sink) {
		    keep = 1;
		} else {
		    if (form == sink
			&& (odbx_data->bits & odbx_text->bits & BACK_LINK)) {
			keep = 1;
		    } else {
			keep = 0;

			if (odbx_text->bits & RAIL_TRIP) {
			    if (((struct e4_2nde_type *)
				 odbx_text->extension2)->bits
				& (RAIL_LIST | RAIL_MARK)) {
				if ((char *) odbx_data - (char *) odbx_text
				    == (offsetof(trip_type, rail)
					- offsetof(trip_type, lead))) {
				    keep = -1;
				} else {
				    keep = 1;
				}
			    }
			}
		    }
		}

		if (keep == 1) {
		    status = push_lock
			(&lock_data, atom_data, none, odbx_data, parser_data);
		    if (status) {
			atom_data++;
			break;
		    }
		} else {
		    struct e4_odb2_type *odb2_data = NULL;
		    struct lock_type *free_lock = NULL;

		    do {
			void *data;

			status = M_LINK
			    (parser_data, &data, sizeof(struct e4_odb2_type));
			if (status) {
			    status = ALLOC_ERROR;
			    break;
			}
			if (odb2_data) {
			    odb2_data->odb2_data = data;
			    odb2_data = data;
			} else {
			    atom_data->odb2._2nd = data;
			    odb2_data = data;
			}

			odb2_data->odb2_data = NULL;

			status = case_odb2
			    (lock_data, &none, &call, parser_data);
			if (status) {
			    break;
			}

			odb2_data->odbx_data = lock_data->odbx_data;

			pick_lock(&lock_data, &free_lock, parser_data);

			if (keep == -1) {
			    if (lock_data) {
				if (lock_data->odbx_data->pass < form) {
				    status = _x1f4_e4_170621_error_2
					(parser_data, lock_data->odbx_data);
				}
			    }

			    break;
			}

			if (lock_data) {
			} else {
			    break;
			}

			odbx_text = lock_data->odbx_data;

			sink = odbx_text->pass;

			if (form < sink) {
			    break;
			}

			if (form == sink
			    && (odbx_data->bits & odbx_text->bits
				& BACK_LINK)) {
			    break;
			}

			if (odbx_text->bits & RAIL_TRIP) {
			    if (((struct e4_2nde_type *)
				 odbx_text->extension2)->bits
				& (RAIL_LIST | RAIL_MARK)) {
				if ((char *) odbx_data - (char *) odbx_text
				    == (offsetof(trip_type, rail)
					- offsetof(trip_type, lead))) {
				    keep = -1;
				} else {
				    break;
				}
			    }
			}
		    } while (1);
		    if (status) {
			atom_data++;

			if (free_lock) {
			    M_FREE(parser_data, free_lock);
			}

			break;
		    }

		    free_lock->lock_data = lock_data;
		    lock_data = free_lock;

		    lock_data->atom_data = NULL;
		    lock_data->odbx_data = odbx_data;
		    lock_data->type = none;
		}
	    }
	}

	atom_data++;
    }

    while (lock_data) {
	pick_lock(&lock_data, NULL, parser_data);
    }

    if (call) {
	if (status) {
	} else {
	    status = call_hack(hack_data, parser_data);
	}
    }

    if (status) {
	hack_data->count =
	    atom_data - (struct e4_atom_type *) hack_data->atoms;
	free_odbx(hack_data, parser_data->expression_data);
	hack_data->count = count;
    } else {
	*type = hack_type;

	_x1f4_e4_fast_hack(hack_data, parser_data->dddevice);
    }

    return status;
}


static int
near_last(struct e4_atom_type *atom_data, struct half_type *half_data,
	  struct parser_type *parser_data)
{
    int status = 0, type;
    const int *args;
    struct e4_hack_type *hack_data;
    const struct e4_last_type *last_data;
    unsigned bits, count, j;

    type = atom_data->type;

    hack_data = atom_data->data.last.base;
    last_data = hack_base(hack_data)->last_data;

    bits = last_data->bits;

    if (0) {
    } else {
	args = last_data->args;
    }

    count = last_data->count;

    j = count;

    while (j) {
	int type;
	unsigned class;

	parser_data->post_ever = NULL;

	status = link_hack(hack_data, &type, parser_data);
	if (status) {
	    break;
	}

	--j;

	if (!(bits & POST_TYPE)) {
	    class = 0;

	    if (type != *args) {
		status = type_last
		    (hack_data, &type, args, j, last_data, atom_data,
		     parser_data);
		if (status) {
		    hack_data++;
		    break;
		}
	    }
	} else {
	    class = args[count];

	    if (!(class & POST_XSET)) {
		if (type != *args) {
		    status = type_last
			(hack_data, &type, args, j, last_data, atom_data,
			 parser_data);
		    if (status) {
			hack_data++;
			break;
		    } else {
			class = args[count];
		    }
		}
	    } else {
		if (hack_data->bits & READ_ONLY) {
		    status = type_post
			(hack_data, j, last_data, atom_data, parser_data);
		    if (status) {
			hack_data++;
			break;
		    }
		} else {
		    if (type != *args) {
			status = type_miss
			    (hack_data, &type, args, j, last_data, atom_data,
			     parser_data);
			if (status) {
			    hack_data++;
			    break;
			}
		    }
		    {
			struct e4_ever_type *ever_data;

			ever_data = (void *) parser_data->post_ever;
			if (ever_data) {
			    if (class & POST_FLAT) {
				ever_data->bits &= ~PRINCIPAL;
			    }
			} else {
			}
		    }
		}
	    }
	}

	_x1f4_e4_fast_miss(hack_data, class);

	hack_data++;

	args++;
    }
    if (status) {
	j = last_data->count - j;
	for (; j; j--) {
	    --hack_data;
	    free_odbx(hack_data, parser_data->expression_data);
	}
    } else {
	*half_data->none = last_data->type;

	if (type == PICK) {
	    atom_data->load = _x1f4_e4_load_pick;
	} else {
	    if (type == LOCK) {
		atom_data->load = _x1f4_e4_load_lock;
	    } else {
		if (1) {
		    if (1) {
			if (last_data->bits & TEXT_LINK) {
			    _x1f4_e4_link_laSt(atom_data);
			} else {
			    _x1f4_e4_link_last(atom_data);
			}
		    }
		}
	    }
	}
    }

    return status;
}


static int
near_mode(struct e4_atom_type *atom_data, struct half_type *half_data,
	  struct parser_type *parser_data)
{
    atom_data->load = _x1f4_e4_load_mode;

    return 0;
}


static int
near_real(struct e4_atom_type *atom_data, struct half_type *half_data,
	  struct parser_type *parser_data)
{
    atom_data->load = _x1f4_e4_load_real;

    return 0;
}


static int
near_text(struct e4_atom_type *atom_data, struct half_type *half_data,
	  struct parser_type *parser_data)
{
    atom_data->load = _x1f4_e4_load_text;

    return 0;
}


static int
near_this(struct e4_atom_type *atom_data, struct half_type *half_data,
	  struct parser_type *parser_data)
{
    int status;

    status = link_hack
	(atom_data->data.hack.data, half_data->none, parser_data);
    if (status) {
    } else {
	if (1) {
	    atom_data->load = _x1f4_e4_load_this;
	}
    }

    return status;
}


static int
pass_type(struct e4_odb1_type *odb1_data, int name, int type)
{
    if (name == BILL) {
	if (type == MODE) {
	    odb1_data->odbx_data = _x1f4_e4_implicit + 0;
	} else {
	    odb1_data->odbx_data = _x1f4_e4_implicit + 1;
	}
    } else {
	if (name == MODE) {
	    if (type == BILL) {
		odb1_data->odbx_data = _x1f4_e4_implicit + 2;
	    } else {
		odb1_data->odbx_data = _x1f4_e4_implicit + 3;
	    }
	} else {
	    if (type == BILL) {
		odb1_data->odbx_data = _x1f4_e4_implicit + 4;
	    } else {
		odb1_data->odbx_data = _x1f4_e4_implicit + 5;
	    }
	}
    }

    return 0;
}


static int
pick_lock(struct lock_type **lock, struct lock_type **free,
	  struct parser_type *parser_data)
{
    struct lock_type *lock_data;

    lock_data = *lock;

    if (free) {
	struct lock_type *free_lock;

	free_lock = *free;
	if (free_lock) {
	    M_FREE(parser_data, free_lock);
	}

	*free = lock_data;
	*lock = lock_data->lock_data;
    } else {
	*lock = lock_data->lock_data;

	M_FREE(parser_data, lock_data);
    }

    return 0;
}


static int
push_lock(struct lock_type **lock, struct e4_atom_type *atom_data, int none,
	  const struct e4_odbx_type *odbx_data,
	  struct parser_type *parser_data)
{
    int status;
    struct lock_type *lock_data;

    status = M_LINK(parser_data, &lock_data, sizeof(struct lock_type));
    if (status) {
	status = ALLOC_ERROR;
    } else {
	lock_data->lock_data = *lock;
	*lock = lock_data;

	lock_data->atom_data = atom_data;
	lock_data->odbx_data = odbx_data;
	lock_data->type = none;
    }

    return status;
}


static int
type_call(struct e4_hack_type *hack_data, int *type, int miss, int name,
	  struct parser_type *parser_data)
{
    int status;

    /*
     * the _load_ method needs be set if hack is modified.
     * _lead_type_ does modify it in certain cases, yet for such cases
     * (single atom, unary operator) the subsequent call to _x1f4_e4_fast_miss_
     * sets the _load_ method.
     */

    do {
	if ((miss | name) & ~15) {
	} else {
	    if ((1 << miss) & SCALAR_BITS && (1 << name) & SCALAR_BITS) {
		status = lead_type(hack_data, miss, name, parser_data, NULL);
		if (status) {
		} else {
		    *type = name;
		}

		break;
	    }
	}

	if (parser_data->bits & IMPLICIT_LINK) {
	    int (*lock)(const void *, int, int, const struct e4_odbx_type **);
	    const struct e4_attributes_type *attributes_data;

	    attributes_data = parser_data->attributes_data;

	    lock = attributes_data->trap_0.get;
	    if (lock) {
		const struct e4_odbx_type *odbx_data;

		if (lock
		    (attributes_data->trap_0.context, name, miss,
		     &odbx_data)) {
		} else {
		    status = lead_type
			(hack_data, miss, name, parser_data, odbx_data);
		    if (status) {
		    } else {
			*type = name;
		    }

		    break;
		}
	    }
	}

	if (parser_data->bits & COMPOSER_LINK) {
	    int (*lock)(const void *, int, int, const struct e4_line_type **);
	    const struct e4_attributes_type *attributes_data;

	    attributes_data = parser_data->attributes_data;

	    lock = attributes_data->turn_0.get;
	    if (lock) {
		const struct e4_line_type *line_data;

		if (lock
		    (attributes_data->turn_0.context, name, miss,
		     &line_data)) {
		} else {
		    struct e4_atom_type *atom_data;

		    /*
		     * TODO
		     *
		     * make a function out of this block.  as for the
		     * IMPLICIT_LINK fix.
		     */
		    status = M_LINK(parser_data, &atom_data,
				    sizeof(struct e4_atom_type));
		    if (status) {
			status = ALLOC_ERROR;
		    } else {
			struct e4_hack_type *hack_this;
			const struct e4_last_type *last_data;
			unsigned offset;

			last_data = &line_data->last;

#define BACK_HACK(count, excess) \
    (offset +           sizeof(struct e4_hack_type))


			offset = sizeof(struct e4_base_type);

			status = M_LINK(parser_data, &hack_this,
					BACK_HACK(count, excess));
			if (status) {
			    status = ALLOC_ERROR;
			    M_FREE(parser_data, atom_data);
			} else {
			    hack_this = (struct e4_hack_type *)
				((char *) hack_this + offset);

			    if (last_data->bits & TEXT_LINK) {
				hack_base(hack_this)->text =
				    (void *) line_data->text;
			    }

			    atom_data->odb1 = NULL;
			    atom_data->odb2._1st = NULL;
			    atom_data->type = LAST;
			    atom_data->data.last.base = hack_this;
			    hack_base(hack_this)->last_data = last_data;

			    _x1f4_e4_fast_miss(hack_data, 0);

			    *hack_this = *hack_data;

			    hack_data->count = 1;
			    hack_data->atoms = atom_data;
			    hack_data->bits = READ_ONLY;

			    _x1f4_e4_fast_hack
				(hack_data, parser_data->dddevice);

			    if (1) {
				if (last_data->bits & TEXT_LINK) {
				    _x1f4_e4_link_laSt(atom_data);
				} else {
				    _x1f4_e4_link_last(atom_data);
				}
			    }

			    *type = name;
			}

#undef BACK_HACK
		    }

		    break;
		}
	    }
	}

	status = PARSE_ERROR;
    } while (0);

    return status;
}


static int
type_hack(struct e4_hack_type *hack_data, int *type, int miss, int name,
	  struct parser_type *parser_data)
{
    int status;

    status = type_call(hack_data, type, miss, name, parser_data);
    if (status) {
	if (status == PARSE_ERROR) {
	    status = _x1f4_e4_130607_error_0(parser_data, name, miss);
	}
    }

    return status;
}


static int
type_last(struct e4_hack_type *hack_data, int *type, const int *args,
	  unsigned j, const struct e4_last_type *last_data,
	  struct e4_atom_type *atom_data, struct parser_type *parser_data)
{
    int miss, name, status;

    miss = *type;
    name = *args;

    if (name == SIDE) {
	if (miss == VOID) {
	    status = _x1f4_e4_130704_error_0(parser_data);
	} else {
	    int *down;

	    down = (int *) args;
	    *down = miss;

	    status = 0;

	    if (hack_data->bits & READ_ONLY) {
	    } else {
		unsigned bits;

		bits = last_data->bits;
		if (bits & LINK_PASS) {
		    if (bits & FLAT_LINE) {
			const struct e4_last_type *last_text;

			last_text = (void *)
			    ((struct e4_flat_type *) last_data - 2)->flat_data;
			if (last_text->count < last_data->count - j) {
			    down[last_data->count] |= POST_XSET;
			}
		    } else {
			/*
			 * not clear how it can get here.
			 */
			down[last_data->count] |= POST_XSET;
		    }
		}
	    }
	}
    } else {
	status = type_call(hack_data, type, miss, name, parser_data);
	if (status) {
	    if (status == PARSE_ERROR) {
		status = _x1f4_e4_080704_error_k
		    (parser_data, atom_data, last_data->count - j, name,
		     miss, parser_data->expression);
	    }
	}
    }

    return status;
}


static int
type_miss(struct e4_hack_type *hack_data, int *type, const int *args,
	  unsigned j, const struct e4_last_type *last_data,
	  struct e4_atom_type *atom_data, struct parser_type *parser_data)
{
    int status;

    if (*args == SIDE) {
	int *down;

	down = (int *) args;
	*down = *type;

	status = 0;
    } else {
	status = _x1f4_e4_081014_error_1
	    (parser_data, hack_data, atom_data, last_data->count - j, *args,
	     *type, parser_data->expression);
    }

    return status;
}


static int
type_post(struct e4_hack_type *hack_data, unsigned j,
	  const struct e4_last_type *last_data, struct e4_atom_type *atom_data,
	  struct parser_type *parser_data)
{
    return _x1f4_e4_081014_error_2
	(parser_data, hack_data, atom_data, last_data->count - j,
	 parser_data->expression);
}


static int
type_trip(int *trip, struct e4_atom_type *atom_data, unsigned lead,
	  struct parser_type *parser_data)
{
    int status = 0;

    if (*trip == SIDE) {
	if (lead == 1) {
	    int type;

	    type = atom_data->type;
	    switch (type) {
	    case BILL:
	    case MODE:
	    case REAL:
	    case TEXT:
		*trip = type;
		break;
	    case EVER:
		*trip = atom_data->data.ever.ever_data->type;
		break;
	    case LAST:
	    case LOCK:
	    case PICK:
		*trip = last_base(atom_data)->last_data->type;
		break;
	    default:
		status = _x1f4_e4_type_hack
		    (atom_data->data.hack.data, trip, parser_data);
	    }
	} else {
	    /* never needed, never tested, should be trivial */
	    status = PARSE_ERROR;
	}
    }

    return status;
}


static void
free_odb1(struct e4_odb1_type *odb1_data, void *expression)
{
    while (odb1_data) {
	struct e4_odb1_type *free_odb1;

	free_odb1 = odb1_data;
	odb1_data = odb1_data->odb1_data;
	E_FREE(expression, free_odb1);
    }
}


static void
free_odbx(struct e4_hack_type *hack_data, void *expression)
{
    struct e4_atom_type *atom_data;
    unsigned i;

    atom_data = hack_data->atoms;
    i = hack_data->count;
    for (; i; i--) {
	_x1f4_e4_free_odbx(atom_data, expression);
	atom_data++;
    }
}


#if __ALLOW_RIGHT_FIRST__
static void
swap_call(struct e4_atom_type *atom_data, unsigned count)
{
    struct e4_atom_type *atom_text;

    atom_text = atom_data + count;
    count >>= 1;
    for (; count; count--) {
	struct e4_atom_type atom;

	atom = *--atom_text;
	*atom_text = *atom_data;
	*atom_data++ = atom;
    }
}
#endif				/* __ALLOW_RIGHT_FIRST__ */


int
_x1f4_e4_link_hack(struct e4_hack_type *hack_data, int *type, void *parser)
{
    int status;

    status = link_hack(hack_data, type, parser);
    if (status) {
    } else {
	struct parser_type *parser_data;

	parser_data = parser;
	if (parser_data->bits & SEQUENCE_LINK) {
	    int miss, name;

	    miss = *type;
	    name = parser_data->attributes_data->link_z.type;
	    if (name ^ miss) {
		status = type_hack(hack_data, type, miss, name, parser_data);
	    }
	}
    }

    return status;
}
