/*
 * e4.7.4.c
 * Copyright (C) 2006-2017, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>
#include <string.h>

#include <e4-defs.h>
#include <e4-inter.h>
#include <e4-parse.h>
#include <e4-types.h>

int
_x1f4_e4_type_risk(const struct e4_odbx_type **odbx, int *lock_type, int major,
		   int minor, const struct e4_odbx_type *odbx_lock,
		   void *parser)
{
    const char *name;
    int rail, status = 0;
    const struct e4_odbx_type *const *list, *odbx_data = NULL;
    unsigned length;

    if (odbx_lock->bits & E2ND_LINK) {
	rail = ((const struct e4_2nde_type *) odbx_lock->extension2)->bits
	    & BACK_RAIL;
    } else {
	rail = 0;
    }

    length = odbx_lock->size;
    name = odbx_lock->name;
    list = ((struct parser_type *) parser)->attributes_data->odb2_data;
    if (rail) {
	const struct e4_2nde_type *e2nd_data;
	const struct e4_odbx_type *odbx_slip = NULL;
	struct trip_type *trip_data;

	trip_data = (void *)
	    ((char *) odbx_lock - offsetof(struct trip_type, rail));

	e2nd_data = &trip_data->left;

	if (e2nd_data->bits & RAIL_MARK) {
	    const int *args;

	    odbx_slip = e2nd_data->rail;
	    if (odbx_slip->size == length) {
		args = odbx_slip->args;
		if (args[1] == minor) {
		    if (memcmp(odbx_slip->name, name, length)) {
		    } else {
			odbx_data = odbx_slip;
		    }
		}
	    }
	} else {
	    const struct e4_rail_type *rail_data;
	    unsigned i;

	    rail_data = e2nd_data->rail;
	    odbx_slip = rail_data->odbx_data;
	    i = rail_data->miss;
	    for (; i; i--) {
		if (odbx_slip->size == length) {
		    const int *args;

		    args = odbx_slip->args;
		    if (args[1] == minor) {
			if (memcmp(odbx_slip->name, name, length)) {
			} else {
			    odbx_data = odbx_slip;
			    if (1) {
				break;
			    }
			}
		    }
		}

		odbx_slip++;
	    }
	}

	if (odbx_data) {
	    trip_data->rail = *odbx_data;
	    trip_data->post = *(struct e4_2nde_type *) odbx_data->extension2;
	    trip_data->post.rail = &trip_data->lead;
	    trip_data->rail.extension2 = &trip_data->post;
	    trip_data->left.bits &= ~CALL_LINK;

	    odbx_data = &trip_data->rail;

	    *lock_type = odbx_data->type;
	} else {
	    status = _x1f4_e4_170621_error_3(parser, trip_data, minor);
	}
    } else {
	if (odbx_lock->bits & RAIL_TRIP) {
	    odbx_data = *list;

	    while (odbx_data) {
		if (odbx_data->bits & E2ND_LINK) {
		    const struct e4_2nde_type *e2nd_data;

		    e2nd_data = odbx_data->extension2;
		    if (e2nd_data->bits & (RAIL_LIST | RAIL_MARK)) {
			if (odbx_data->size == length) {
			    const int *args;

			    args = odbx_data->args;
			    if (args[0] == major) {
				if (args[1] == minor) {
				    if (memcmp
					(odbx_data->name, name, length)) {
				    } else {
					*lock_type = odbx_data->type;
					break;
				    }
				}
			    }
			}
		    }
		}

		list++;
		odbx_data = *list;
	    }

	    if (odbx_data) {
		struct e4_2nde_type *e2nd_trip;
		struct e4_odbx_type *odbx_trip;

		odbx_trip = (struct e4_odbx_type *) odbx_lock;
		e2nd_trip = (struct e4_2nde_type *) odbx_trip->extension2;
		*odbx_trip = *odbx_data;
		*e2nd_trip = *(struct e4_2nde_type *) odbx_data->extension2;
		odbx_data = odbx_trip;
		odbx_trip->bits |= RAIL_TRIP;
		e2nd_trip->bits &= ~CALL_LINK;
		odbx_trip->extension2 = e2nd_trip;
	    } else {
		status = _x1f4_e4_170621_error_4
		    (parser,
		     (char *) odbx_lock - offsetof(struct trip_type, lead),
		     major, minor);
	    }
	} else {
	    odbx_data = *list;

	    while (odbx_data) {
		unsigned size;

		size = odbx_data->size;

		if (odbx_data->bits & E2ND_LINK) {
		    if (((struct e4_2nde_type *) odbx_data->extension2)->bits
			& RAIL_ONLY) {
			size = 0;
		    }
		}

		if (size == length) {
		    const int *args;

		    args = odbx_data->args;
		    if (args[0] == major) {
			if (args[1] == minor) {
			    if (memcmp(odbx_data->name, name, length)) {
			    } else {
				*lock_type = odbx_data->type;
				break;
			    }
			}
		    }
		    if ((odbx_data->bits & (CALL_LOCK | LEFT_XSET))
			== (CALL_LOCK | LEFT_XSET) && BASE_LOCK < major) {
			if ((BASE_LOCK < minor ? major : args[1]) == minor) {
			    if (memcmp(odbx_data->name, name, length)) {
			    } else {
				*lock_type = major;
				break;
			    }
			} else {
			}
		    }
		}

		list++;
		odbx_data = *list;
	    }

	    if (odbx_data) {
	    } else {
		status = _x1f4_e4_080704_error_6
		    (parser, name, length, major, minor);
	    }
	}
    }

    if (odbx_data) {
	*odbx = odbx_data;
    }

    return status;
}
