/*
 * lxdash-s.c.c
 * Copyright (C) 2008-2015, 2017, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>

#include <d_fset.h>
#include <e4.h>
#include <fd.h>
#include <lxdash-inter.h>
#include <lxdash-types.h>

#define I_LINE(t)			(*((X1f4_E4_C_USER *) (t)))
#define I_MODE(t)			(*((X1f4_E4_C_MODE *) (t)))
#define I_USER(t)			(*((X1f4_E4_C_USER *) (t)))

#define l_USER(e, output) \
    {									      \
	X1f4_E4_C_USER *l;						      \
									      \
	l = (void *) (output);						      \
	*l = (e);							      \
    }

int
_libx1f4i0_lxdash_c_sequent(void *context, void *output, void **input)
{
    int status;

    do {
	struct dxdash_type *dxdash_data;
	struct effect_type effect, *effect_data;
	struct lxdash_type *lxdash_data;
	struct lxtext_type *lxtext_data, *lxtext_link;
	void *data, *side;

	lxtext_data = context;

	dxdash_data = I_LINE(input[0]);

	lxdash_data = lxtext_data->text;

	lxtext_link = lxdash_data->link_f.text;

	effect.mode = I_MODE(input[1]);

	status = x1f4_post_lxfset(dxdash_data->dash, &effect, &data);
	if (status) {
	    if (status ^ X1f4_BQFSET_EVER_MATCH) {
		status = _libx1f4i0_lxdash_stat_line(lxdash_data, 0, status);
		if (1) {
		    break;
		}
	    } else {
		unsigned link;

		effect_data = data;

		link = effect_data->link;

		if (~link) {
		    int (*free)(void *, void *, struct excase_type **,
				const struct excase_type *);
		    struct lxtext_type *lxtext_slip;

		    lxtext_slip = lxtext_link + link;

		    free = lxtext_slip->lxtype.free;
		    if (free) {
			status = free
			    (lxtext_slip->lxtype.context,
			     I_USER(&effect_data->fxdata), NULL,
			     lock_excase(dxdash_data));
			if (status) {
			    break;
			}
		    }
		}
	    }
	} else {
	    effect_data = data;

	    dxdash_data->size++;

	    effect_data->mode = effect.mode;
	}

	{
	    unsigned link, type;

	    type = lxtext_data->lxtype.code;

	    effect_data->type = type;

	    link = lxtext_data - lxtext_link;

	    status = lxtext_data->lxtype.node
		(lxtext_data->lxtype.context, &side,
		 I_USER(input[2]), lock_excase(dxdash_data));
	    if (status) {
		link = ~0;
	    }

	    effect_data->link = link;

	    if (1) {
		l_USER(side, &effect_data->fxdata);
		l_USER(side, output);
	    }
	}
    } while (0);

    return status;
}
