/*
 * lxlist-k.i.c
 * Copyright (C) 2008-2011, 2013-2017, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <copy.h>
#include <e4.h>
#include <fd.h>
#include <l_list.h>
#include <lxlist-drive.h>
#include <lxlist-inter.h>
#include <lxlist-types.h>

#define I_LIST(t)			(*((X1f4_E4_C_USER *) (t)))
#define I_MODE(t)			(*((X1f4_E4_C_MODE *) (t)))
#define I_USER(t)			(*((X1f4_E4_C_USER *) (t)))

int
_libx1f4i0_lxlist_i_keeping(void *context, void *output, void **input)
{
    int status;
    struct cxbend_type *cxbend_data;
    struct dxlist_type *dxlist_data;
    struct lxlist_type *lxlist_data;
    unsigned size;

    cxbend_data = context;

    lxlist_data = cxbend_data->list;

    dxlist_data = I_LIST(input[1]);

    size = dxlist_data->size;
    if (size) {
	void *data, *list;

	list = dxlist_data->list;

	status = cxbend_data->pick(list, &data);
	if (status) {
	    status = _libx1f4i0_lxlist_stat_list(lxlist_data, 0, status);
	} else {
	    unsigned call, type;

	    type = effect_type(data);
	    call = *((struct x1f4_function_type *)
		     ((void **) input[-1])[X1f4_E4_RENDITION])->args;
	    if (type != call) {
		status = _libx1f4i0_lxlist_stat_ends
		    (lxlist_data, cxbend_data->bend, type, call);
	    } else {
		unsigned link;

		link = effect_link(data);

		do {
		    if (~link) {
			struct lxtext_type *lxtext_data;

			lxtext_data = lxlist_data->link_f.text;
			lxtext_data += link;
			if (lxtext_data->lxtype.push) {
			    status = lxtext_data->lxtype.push
				(lxtext_data->lxtype.context,
				 *input, I_USER(byte(data) + 8),
				 lock_excase(dxlist_data));
			    if (status) {
				break;
			    }
			} else {
			    copy_miss(*input, type, byte(data) + 8);
			}

			byte(data)[4] = ~0 >> 030;
			byte(data)[5] = ~0 >> 020;
			byte(data)[6] = ~0 >> 010;
			byte(data)[7] = ~0 >> 000;
		    } else {
			copy_miss(*input, type, byte(data) + 8);
		    }

		    status = cxbend_data->slip(list);
		    if (status) {
			status = _libx1f4i0_lxlist_stat_list
			    (lxlist_data, 0, status);
		    } else {
			dxlist_data->size = size - 1;
		    }
		} while (0);
	    }
	}
    } else {
	status = _libx1f4i0_lxlist_stat_null(lxlist_data);
    }

    return status;
}
