/*
 * lxport-types.h
 * Copyright (C) 2008-2016, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __LIBx1f4_lxport_types_H__
#define __LIBx1f4_lxport_types_H__

#include <config.h>

#include <lxport-config.h>

#include <sys/types.h>

#include <dt.h>
#include <frame.h>
#include <integral_q.h>
#include <ln.h>
#include <lxpost.h>
#include <nt.h>
#include <track.h>
#include <trans.h>

#define here_NULL			((void *) 0)

#define here_offsetof(type, ceil) \
    ((integral_q) (&((type *) here_NULL)->ceil))

#define lock_excase(dxlist_text) \
    ((struct x1f4_nodelink_type *) &lock_cxlist((dxlist_text))->link)

#define lock_exlist(dxlist_text) \
    ((struct x1f4_nodeline_type *) &lock_cxlist((dxlist_text))->line)

#define lock_cxlist(dxlist_text) \
    ((struct x1f4_leadnode_type *)					      \
     ((char *) (dxlist_text) - sizeof(struct x1f4_leadnode_type)))

#define near_dxlist(excase_text) \
    ((struct ssport_type *)						      \
     ((char *) near_cxlist((excase_text)) + sizeof(struct x1f4_leadnode_type)))

#define near_vxlist(excase_text) \
    ((struct xsport_type *)						      \
     ((char *) near_cxlist((excase_text)) + sizeof(struct x1f4_leadnode_type)))

#define near_exlist(excase_text) \
    ((struct x1f4_nodeline_type *) &near_cxlist((excase_text))->line)

#define near_cxlist(excase_text) \
    ((struct x1f4_leadnode_type *)					      \
     ((char *) (excase_text) - here_offsetof(struct x1f4_leadnode_type, link)))

#define pick_dxlist(fxlist_text) \
    ((struct ssport_type *)						      \
     ((char *) (fxlist_text) + sizeof(struct x1f4_leadnode_type)))

#define pick_vxlist(fxlist_text) \
    ((struct xsport_type *)						      \
     ((char *) (fxlist_text) + sizeof(struct x1f4_leadnode_type)))

#define __struct_ssport_type_alloc_size__ \
    sizeof(struct x1f4_leadnode_type) + sizeof(struct ssport_type)

#define __struct_xsport_type_alloc_size__ \
    sizeof(struct x1f4_leadnode_type) + sizeof(struct xsport_type)

/*
 * _lose_port assumes this belongs to an xsport object.
 */
typedef struct waitlink_type {
    char **argv, *path;
    int status;
    pid_t port;
    struct x1f4_nodelink_type nodelink;
    struct waitlink_type *waitlink_last, *waitlink_next;
    void *xsport;
} waitlink_type;

typedef struct ssport_type {
    void *argv, *path, *read, *wait;
} ssport_type;

typedef struct xsport_type {
    struct waitlink_type *waitlink_data;
    unsigned bits;
    void *argv, *fade, *line, *note, *path, *side, *text;
} xsport_type;

typedef struct lxport_type {
    struct lxpost_type ever_p;
    struct {
	void *data, *side;
    } link_f;
    struct {
	int data, list;
    } link_i;
    struct {
	void *data, *date, *file, *list;
    } link_l;
    struct {
	int (*call)(void *, struct x1f4_nodelink_type *),
	    (*fine)(void *, struct x1f4_nodelink_type **,
		    struct x1f4_nodelink_type *),
	    (*miss)(void *, struct x1f4_nodelink_type **,
		    struct x1f4_nodelink_type *);
	void *text;
    } link_r;
    struct {
	void *wait;
    } link_u;
    struct {
	int c3rd;
    } link_x;
    struct {
	int (*call)(void *);
	void *back;
    } link_z;
} lxport_type;

typedef struct screen_type {
    struct x1f4_track_type lock_e;
    struct {
	int data, file, list, time;
    } link_i;
    struct {
	int ssxx, xsxx;
    } link_l;
    struct x1f4_trans_type lock_m;
    struct {
	int (*call)(void *, struct x1f4_nodelink_type *),
	    (*fine)(void *, struct x1f4_nodelink_type **,
		    struct x1f4_nodelink_type *),
	    (*miss)(void *, struct x1f4_nodelink_type **,
		    struct x1f4_nodelink_type *);
	void *text;
    } link_r;
    struct x1f4_frame_type lock_v;
    struct x1f4_trans_type lock_w;
    struct {
	int c3rd;
    } link_x;
    struct {
	int (*call)(void *);
	void *back;
    } link_z;
} screen_type;

#endif				/* __LIBx1f4_lxport_types_H__ */
